@@ -47,6 +47,8 @@
 #include <QMap>
 #include <QHash>
 #include <QXmlStreamEntityResolver>
+#include <QPointer>
+#include <QMutableListIterator>
 
 #include <kconfiggroup.h>
 #include <KIconThemes/KIconLoader>
@@ -63,7 +65,7 @@ public:
      * @internal
      * Fills the KPassivePopup with data
      */
-    void fillPopup(KPassivePopup *popup, KNotification *notification, KNotifyConfig *config);
+    void fillPopup(KPassivePopup *popup, KNotification *notification, const KNotifyConfig &config);
     /**
      * Make sure a popup is completely supported by the notification backend.
      * Changes the popup to be compatible if needed.
@@ -86,7 +88,7 @@ public:
      *               Otherwise will put new notification on screen
      * @return true for success or false if there was an error.
      */
-    bool sendNotificationToGalagoServer(KNotification *notification, KNotifyConfig *config, bool update = false);
+    bool sendNotificationToGalagoServer(KNotification *notification, const KNotifyConfig &config, bool update = false);
     /**
      * Sends request to close Notification with id to DBus "org.freedesktop.notifications" interface
      *  @param id knotify-side notification ID to close
@@ -95,7 +97,7 @@ public:
     /**
      * Find the caption and the icon name of the application
      */
-    void getAppCaptionAndIconName(KNotifyConfig *config, QString *appCaption, QString *iconName);
+    void getAppCaptionAndIconName(const KNotifyConfig &config, QString *appCaption, QString *iconName);
     /*
      * Query the dbus server for notification capabilities
      * If no DBus server is present, use fallback capabilities for KPassivePopup
@@ -122,7 +124,7 @@ public:
      * so we queue all notifications while waiting for the capabilities
      * to return, then process them from this queue
      */
-    QList<QPair<KNotification*, KNotifyConfig*> > notificationQueue;
+    QList<QPair<KNotification*, KNotifyConfig> > notificationQueue;
     /**
      * Whether the DBus notification daemon capability cache is up-to-date.
      */
@@ -137,7 +139,8 @@ public:
      * As we communicate with the notification server over dbus
      * we use only ids, this is for fast KNotifications lookup
      */
-    QHash<uint, KNotification*> galagoNotifications;
+    QHash<uint, QPointer<KNotification>> galagoNotifications;
+
 
     NotifyByPopup * const q;
 
@@ -147,14 +150,14 @@ public:
      */
     class HtmlEntityResolver : public QXmlStreamEntityResolver
     {
-        QString resolveUndeclaredEntity(const QString &name);
+        QString resolveUndeclaredEntity(const QString &name) Q_DECL_OVERRIDE;
     };
 
 };
 
 //---------------------------------------------------------------------------------------
 
-NotifyByPopup::NotifyByPopup(QObject *parent) 
+NotifyByPopup::NotifyByPopup(QObject *parent)
   : KNotificationPlugin(parent),
     d(new NotifyByPopupPrivate(this))
 {
@@ -222,6 +225,11 @@ NotifyByPopup::~NotifyByPopup()
 }
 
 void NotifyByPopup::notify(KNotification *notification, KNotifyConfig *notifyConfig)
+{
+    notify(notification, *notifyConfig);
+}
+
+void NotifyByPopup::notify(KNotification *notification, const KNotifyConfig &notifyConfig)
 {
     if (d->passivePopups.contains(notification) || d->galagoNotifications.contains(notification->id())) {
         // notification is alrady on the screen, do nothing
@@ -268,7 +276,6 @@ void NotifyByPopup::notify(KNotification *notification, KNotifyConfig *notifyCon
 
         // Finish immediately, because current NotifyByPopupGrowl can't callback
         finish(notification);
-        delete notifyConfig;
         return;
     }
 
@@ -363,9 +370,22 @@ void NotifyByPopup::close(KNotification *notification)
         // which will call finish() on the notification
         d->passivePopups[notification]->deleteLater();
     }
+
+    QMutableListIterator<QPair<KNotification*, KNotifyConfig> > iter(d->notificationQueue);
+    while (iter.hasNext()) {
+        auto &item = iter.next();
+        if (item.first == notification) {
+            iter.remove();
+        }
+    }
 }
 
 void NotifyByPopup::update(KNotification *notification, KNotifyConfig *notifyConfig)
+{
+    update(notification, *notifyConfig);
+}
+
+void NotifyByPopup::update(KNotification *notification, const KNotifyConfig &notifyConfig)
 {
     if (d->passivePopups.contains(notification)) {
         KPassivePopup *p = d->passivePopups[notification];
@@ -388,9 +408,12 @@ void NotifyByPopup::update(KNotification *notification, KNotifyConfig *notifyCon
 
 void NotifyByPopup::onServiceOwnerChanged(const QString &serviceName, const QString &oldOwner, const QString &newOwner)
 {
+    Q_UNUSED(serviceName);
     // close all notifications we currently hold reference to
     Q_FOREACH (KNotification *n, d->galagoNotifications.values()) {
-        finished(n);
+        if (n) {
+            finished(n);
+        }
     }
     Q_FOREACH (KNotification *n, d->passivePopups.keys()) {
         finished(n);
@@ -432,38 +455,49 @@ void NotifyByPopup::onServiceOwnerChanged(const QString &serviceName, const QStr
 
 void NotifyByPopup::onGalagoNotificationActionInvoked(uint notificationId, const QString &actionKey)
 {
-    KNotification *n = d->galagoNotifications[notificationId];
-    if (n == 0) {
+    auto iter = d->galagoNotifications.find(notificationId);
+    if (iter == d->galagoNotifications.end()) {
         qWarning() << "Failed to find KNotification id for dbus_id" << notificationId << "- action not triggered";
         return;
     }
 
-    emit actionInvoked(n->id(), actionKey.toUInt());
-    // now close notification - similar to popup behaviour
-    // (popups are hidden after link activation - see 'connects' of linkActivated signal above)
-    d->closeGalagoNotification(n);
+    KNotification *n = *iter;
+    if (n) {
+        emit actionInvoked(n->id(), actionKey.toUInt());
+        // now close notification - similar to popup behaviour
+        // (popups are hidden after link activation - see 'connects' of linkActivated signal above)
+        d->closeGalagoNotification(n);
+    } else {
+        d->galagoNotifications.erase(iter);
+    }
 }
 
 void NotifyByPopup::onGalagoNotificationClosed(uint dbus_id, uint reason)
 {
     Q_UNUSED(reason)
-    KNotification *n = d->galagoNotifications[dbus_id];
-    if (n == 0) {
+    auto iter = d->galagoNotifications.find(dbus_id);
+    if (iter == d->galagoNotifications.end()) {
         qWarning() << "Failed to find KNotification for dbus_id" << dbus_id;
         return;
     }
+    KNotification *n = *iter;
     d->galagoNotifications.remove(dbus_id);
-    finished(n);
-    // The popup bubble is the only user facing part of a notification,
-    // if the user closes the popup, it means he wants to get rid
-    // of the notification completely, including playing sound etc
-    // Therefore we close the KNotification completely after closing
-    // the popup
-    n->close();
+
+    if (n) {
+        finished(n);
+        // The popup bubble is the only user facing part of a notification,
+        // if the user closes the popup, it means he wants to get rid
+        // of the notification completely, including playing sound etc
+        // Therefore we close the KNotification completely after closing
+        // the popup
+        n->close();
+    }
 }
 
 void NotifyByPopup::onGalagoServerReply(QDBusPendingCallWatcher *watcher)
 {
+    // call deleteLater first, since we might return in the middle of the function
+    watcher->deleteLater();
     KNotification *notification = watcher->property("notificationObject").value<KNotification*>();
     if (!notification) {
         qWarning() << "Invalid notification object passed in DBus reply watcher; notification will probably break";
@@ -473,7 +507,6 @@ void NotifyByPopup::onGalagoServerReply(QDBusPendingCallWatcher *watcher)
     QDBusPendingReply<uint> reply = *watcher;
 
     d->galagoNotifications.insert(reply.argumentAt<0>(), notification);
-    watcher->deleteLater();
 }
 
 void NotifyByPopup::onGalagoServerCapabilitiesReceived(const QStringList &capabilities)
@@ -489,20 +522,20 @@ void NotifyByPopup::onGalagoServerCapabilitiesReceived(const QStringList &capabi
     d->notificationQueue.clear();
 }
 
-void NotifyByPopupPrivate::getAppCaptionAndIconName(KNotifyConfig *notifyConfig, QString *appCaption, QString *iconName)
+void NotifyByPopupPrivate::getAppCaptionAndIconName(const KNotifyConfig &notifyConfig, QString *appCaption, QString *iconName)
 {
-    KConfigGroup globalgroup(&(*notifyConfig->eventsfile), QString("Global"));
-    *appCaption = globalgroup.readEntry("Name", globalgroup.readEntry("Comment", notifyConfig->appname));
+    KConfigGroup globalgroup(&(*notifyConfig.eventsfile), QString("Global"));
+    *appCaption = globalgroup.readEntry("Name", globalgroup.readEntry("Comment", notifyConfig.appname));
 
-    KConfigGroup eventGroup(&(*notifyConfig->eventsfile), QString("Event/%1").arg(notifyConfig->eventid));
+    KConfigGroup eventGroup(&(*notifyConfig.eventsfile), QString("Event/%1").arg(notifyConfig.eventid));
     if (eventGroup.hasKey("IconName")) {
-        *iconName = eventGroup.readEntry("IconName", notifyConfig->appname);
+        *iconName = eventGroup.readEntry("IconName", notifyConfig.appname);
     } else {
-        *iconName = globalgroup.readEntry("IconName", notifyConfig->appname);
+        *iconName = globalgroup.readEntry("IconName", notifyConfig.appname);
     }
 }
 
-void NotifyByPopupPrivate::fillPopup(KPassivePopup *popup, KNotification *notification, KNotifyConfig *notifyConfig)
+void NotifyByPopupPrivate::fillPopup(KPassivePopup *popup, KNotification *notification, const KNotifyConfig &notifyConfig)
 {
     QString appCaption;
     QString iconName;
@@ -573,7 +606,7 @@ void NotifyByPopupPrivate::fillPopup(KPassivePopup *popup, KNotification *notifi
     popup->setView( vb );
 }
 
-bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notification, KNotifyConfig *notifyConfig_nocheck, bool update)
+bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notification, const KNotifyConfig &notifyConfig_nocheck, bool update)
 {
     uint updateId = galagoNotifications.key(notification, 0);
 
@@ -652,7 +685,8 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
 
     QDBusPendingCall notificationCall = QDBusConnection::sessionBus().asyncCall(dbusNotificationMessage, -1);
 
-    QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(notificationCall, q);
+    //parent is set to the notification so that no-one ever accesses a dangling pointer on the notificationObject property
+    QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(notificationCall, notification);
     watcher->setProperty("notificationObject", QVariant::fromValue<KNotification*>(notification));
 
     QObject::connect(watcher, SIGNAL(finished(QDBusPendingCallWatcher*)),
@@ -774,4 +808,3 @@ QString NotifyByPopupPrivate::HtmlEntityResolver::resolveUndeclaredEntity(const
     return QString(ent);
 }
 
-#include "notifybypopup.moc"
